from __future__ import annotations
import math
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from vidhash.video_hash import VideoHash


@dataclass(eq=True, frozen=True)  # type: ignore[misc]
class MatchOptions(ABC):
    hamming_dist: int = 0

    @abstractmethod
    def check_match(self, hash1: VideoHash, hash2: VideoHash) -> bool:
        pass


def _has_overlap(hash1: VideoHash, hash2: VideoHash, required_overlap: float, hamming_distance: int) -> bool:
    overlaps = 0
    for image_hash in hash1.hash_set:
        if hash2.contains_hash(image_hash, hamming_distance):
            overlaps += 1
            if overlaps >= required_overlap:
                return True
    return False


@dataclass(eq=True, frozen=True)
class PercentageMatch(MatchOptions):
    percentage_overlap: float = 50
    ignore_blank: bool = True

    def check_match(self, hash1: VideoHash, hash2: VideoHash) -> bool:
        required_overlap = self.percentage_overlap * min(len(hash1.hash_set), len(hash2.hash_set)) / 100
        return _has_overlap(hash1, hash2, required_overlap, self.hamming_dist)


@dataclass(eq=True, frozen=True)
class AbsoluteMatch(MatchOptions):
    count_overlap: int = 1
    ignore_blank: bool = True

    def check_match(self, hash1: VideoHash, hash2: VideoHash) -> bool:
        required_overlap = min(self.count_overlap, len(hash1.hash_set), len(hash2.hash_set))
        return _has_overlap(hash1, hash2, required_overlap, self.hamming_dist)


@dataclass(eq=True, frozen=True)
class DurationMatch(MatchOptions):
    time_overlap: float = 0

    def _check_match_from(
        self,
        hash1: VideoHash,
        start_frame1: int,
        hash2: VideoHash,
        start_frame2: int,
        target_length: int,
    ) -> bool:
        match_count = 0
        remaining_frames1 = hash1.image_hashes[start_frame1:]
        remaining_frames2 = hash2.image_hashes[start_frame2:]
        if len(remaining_frames1) < target_length or len(remaining_frames2) < target_length:
            return False
        for frame1, frame2 in zip(remaining_frames1, remaining_frames2):
            if (frame1 - frame2) > self.hamming_dist:
                return False
            match_count += 1
            if match_count >= target_length:
                return True
        return match_count >= target_length

    def check_match(self, hash1: VideoHash, hash2: VideoHash) -> bool:
        frame_count = min(
            math.ceil(hash1.hash_options.fps * self.time_overlap), len(hash1.image_hashes), len(hash2.image_hashes)
        )
        shorter, longer = (hash1, hash2) if hash1.video_length < hash2.video_length else (hash2, hash1)
        for frame_num1 in range(len(shorter.image_hashes)):
            for frame_num2 in range(len(longer.image_hashes)):
                if self._check_match_from(shorter, frame_num1, longer, frame_num2, frame_count):
                    return True
        return False



from typing import Callable, Optional, Tuple, List

import numpy as np
import numpy.typing as npt
from PIL.Image import Image


class ImageHash:
    hash: npt.NDArray[np.bool_]
    def __init__(self, binary_array: npt.NDArray[np.bool_]) -> None: ...
    def __sub__(self, other: ImageHash) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __len__(self) -> int: ...

def hex_to_hash(hexstr: str) -> ImageHash: ...
def hex_to_flathash(hexstr: str, hashsize: int) -> ImageHash: ...
def old_hex_to_hash(hexstr: str, hash_size: int = ...) -> ImageHash: ...
def average_hash(image: Image, hash_size: int = ..., mean: Callable[[npt.NDArray], float] = ...) -> ImageHash: ...
def phash(image: Image, hash_size: int = ..., highfreq_factor: int = ...) -> ImageHash: ...
def phash_simple(image: Image, hash_size: int = ..., highfreq_factor: int = ...) -> ImageHash: ...
def dhash(image: Image, hash_size: int = ...) -> ImageHash: ...
def dhash_vertical(image: Image, hash_size: int = ...) -> ImageHash: ...
def whash(image: Image, hash_size: int = ..., image_scale: Optional[int] = ..., mode: str = ..., remove_max_haar_ll: bool = ...) -> ImageHash: ...
def colorhash(image: Image, binbits: int = ...) -> ImageHash: ...

class ImageMultiHash:
    segment_hashes: List[ImageHash]
    def __init__(self, hashes: List[ImageHash]) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __sub__(self, other: ImageMultiHash, hamming_cutoff: Optional[int] = ..., bit_error_rate: Optional[float] = ...) -> int: ...
    def __hash__(self) -> int: ...
    def hash_diff(self, other_hash: ImageMultiHash, hamming_cutoff: Optional[int] = ..., bit_error_rate: Optional[float] = ...) -> Tuple[int, int]: ...
    def matches(self, other_hash: ImageMultiHash, region_cutoff: int = ..., hamming_cutoff: Optional[int] = ..., bit_error_rate: Optional[float] = ...) -> bool: ...
    def best_match(self, other_hashes: List[ImageMultiHash], hamming_cutoff: Optional[int] = ..., bit_error_rate: Optional[float] = ...) -> ImageMultiHash: ...

def crop_resistant_hash(image: Image, hash_func: Callable[[Image], ImageHash] = ..., limit_segments: Optional[int] = ..., segment_threshold: int = ..., min_segment_size: int = ..., segmentation_image_size: int = ...) -> ImageMultiHash: ...



from vidhash.func import hash_video, check_match
from vidhash.hash_options import HashSettings, HashOptions
from vidhash.match_options import MatchOptions
from vidhash.video_hash import VideoHash

__all__ = ["hash_video", "check_match", "HashSettings", "HashOptions", "MatchOptions", "VideoHash"]



from __future__ import annotations

import glob
import os
import pathlib
import shutil
import subprocess
import uuid
from dataclasses import dataclass
from multiprocessing.pool import ThreadPool
from typing import TYPE_CHECKING

import ffmpy3
from PIL import Image

from vidhash.hash_options import HashOptions
from vidhash.match_options import PercentageMatch
from vidhash.video_hash import VideoHash

if TYPE_CHECKING:
    from typing import Tuple, Optional, Dict, List
    from vidhash.match_options import MatchOptions

TEMP_DIR = "vidhash_temp/"
DEFAULT_HASH_OPTS = HashOptions()
DEFAULT_MATCH_OPTS = PercentageMatch(3, 20)


async def _process_ffmpeg(ff: ffmpy3.FFmpeg) -> Tuple[str, str]:
    ff_process = await ff.run_async(stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out_bytes, err_bytes = await ff_process.communicate()
    await ff.wait()
    output = out_bytes.decode("utf-8", errors="replace").strip()
    error = err_bytes.decode("utf-8", errors="replace").strip()
    return output, error


async def _run_ffmpeg(
    inputs: Dict[str, Optional[str]],
    outputs: Dict[str, Optional[str]],
    global_options: Optional[List[str]] = None,
) -> Tuple[str, str]:
    ff = ffmpy3.FFmpeg(global_options=global_options, inputs=inputs, outputs=outputs)
    return await _process_ffmpeg(ff)


async def _run_ffprobe(inputs: Dict[str, Optional[str]], global_options: Optional[List[str]] = None) -> Tuple[str, str]:
    ff = ffmpy3.FFprobe(global_options=global_options, inputs=inputs)
    return await _process_ffmpeg(ff)


def _cleanup_file(path: str) -> None:
    try:
        os.remove(path)
    except FileNotFoundError:
        pass


def _cleanup_dir(path: str) -> None:
    try:
        shutil.rmtree(path)
    except FileNotFoundError:
        pass


async def _decompose_video(video_path: str, decompose_path: str, fps: float, max_size: float) -> None:
    # Convert video and downscale
    output_path = str(pathlib.Path(TEMP_DIR) / f"{uuid.uuid4()}.mp4")
    filters = [
        f"scale='min({max_size},iw)':'min({max_size},ih)':force_original_aspect_ratio=decrease",
        "scale=trunc(iw/2)*2:trunc(ih/2)*2",
    ]
    os.makedirs(TEMP_DIR, exist_ok=True)
    try:
        await _run_ffmpeg(
            inputs={video_path: None},
            outputs={output_path: f"-vf \"{','.join(filters)}\""},
        )
        # Decompose it
        os.makedirs(decompose_path, exist_ok=True)
        await _run_ffmpeg(
            inputs={output_path: None},
            outputs={f"{decompose_path}/out%d.png": f"-vf fps={fps} -vsync 0"},
            global_options=["-y"],
        )
    finally:
        # Clean up the temporary video file
        _cleanup_file(output_path)


async def _video_length(video_path: str) -> float:
    out, err = await _run_ffprobe(
        inputs={video_path: "-show_entries format=duration -of default=noprint_wrappers=1:nokey=1"},
        global_options=["-v error"],
    )
    return float(out)


async def hash_video(video_path: str, hash_options: HashOptions = None) -> VideoHash:
    options = hash_options or DEFAULT_HASH_OPTS
    # Get video length
    video_length = await _video_length(video_path)
    # Decompose into images
    video_id = str(uuid.uuid4())
    decompose_path = str(pathlib.Path(TEMP_DIR) / video_id)
    try:
        await _decompose_video(video_path, decompose_path, options.fps, options.settings.video_size)
        # Hash images
        image_files = glob.glob(f"{decompose_path}/*.png")
        hash_pool = ThreadPool(os.cpu_count())
        hash_list = hash_pool.map(lambda image_path: options.settings.hash_image(Image.open(image_path)), image_files)
    finally:
        _cleanup_dir(decompose_path)
    # Create VideoHash and return
    return VideoHash(hash_list, video_length, options)


@dataclass(eq=True, frozen=True)
class CheckOptions:
    hash_options: HashOptions = DEFAULT_HASH_OPTS
    match_options: MatchOptions = DEFAULT_MATCH_OPTS


async def check_match(video_path_1: str, video_path_2: str, options: CheckOptions = CheckOptions()) -> bool:
    hash1 = await hash_video(video_path_1, options.hash_options)
    hash2 = await hash_video(video_path_2, options.hash_options)
    return options.match_options.check_match(hash1, hash2)



import subprocess
from asyncio.subprocess import Process
from typing import Optional, List, Dict, Union, IO, Tuple, AnyStr


class FFmpeg:
    executable: str
    cmd: str
    process: Optional[subprocess.Popen]
    def __init__(self, executable: str = ..., global_options: Optional[Union[str, List[str]]] = ..., inputs: Optional[Dict[str, Optional[str]]] = ..., outputs: Optional[Dict[str, Optional[str]]] = ...) -> None: ...
    def run(self, input_data: Optional[bytes] = ..., stdout: Optional[IO[AnyStr]] = ..., stderr: Optional[IO[AnyStr]] = ...) -> Tuple[bytes, bytes]: ...
    async def run_async(self, input_data: Optional[bytes] = ..., stdout: Optional[Union[int, IO[AnyStr]]] = ..., stderr: Optional[Union[int, IO[AnyStr]]] = ...) -> Process: ...
    async def wait(self) -> Optional[int]: ...

class FFprobe(FFmpeg):
    def __init__(self, executable: str = ..., global_options: Optional[Union[str, List[str]]] = ..., inputs: Optional[Dict[str, Optional[str]]] = ...) -> None: ...

class FFExecutableNotFoundError(Exception): ...

class FFRuntimeError(Exception):
    cmd: str
    exit_code: int
    stdout: bytes
    stderr: bytes
    def __init__(self, cmd: str, exit_code: int, stdout: Optional[bytes], stderr: Optional[bytes]) -> None: ...



from __future__ import annotations
import dataclasses
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from typing import List, Set, Iterator
    import imagehash
    from vidhash.hash_options import HashOptions


@dataclasses.dataclass
class VideoHash:
    image_hashes: List[imagehash.ImageHash]
    video_length: float
    hash_options: HashOptions

    @property
    def hash_set(self) -> Set[imagehash.ImageHash]:
        return set(self.image_hashes)

    def matching_hashes(self, other_hash: imagehash.ImageHash, hamming_dist: int = 0) -> Iterator[imagehash.ImageHash]:
        for image_hash in self.hash_set:
            if (image_hash - other_hash) <= hamming_dist:
                yield image_hash

    def contains_hash(self, other_hash: imagehash.ImageHash, hamming_dist: int = 0) -> bool:
        return any(self.matching_hashes(other_hash, hamming_dist))



from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING

import imagehash
import numpy as np

if TYPE_CHECKING:
    from PIL.Image import Image


class HashSettings(ABC):
    @property
    @abstractmethod
    def video_size(self) -> int:
        pass

    @property
    @abstractmethod
    def blank_hash(self) -> imagehash.ImageHash:
        pass

    @abstractmethod
    def hash_image(self, img: Image) -> imagehash.ImageHash:
        pass


@dataclass(eq=True, frozen=True)
class DHash(HashSettings):
    hash_size: int = 8

    def hash_image(self, img: Image) -> imagehash.ImageHash:
        return imagehash.dhash(img, hash_size=self.hash_size)

    @property
    def video_size(self) -> int:
        return self.hash_size * 10

    @property
    def blank_hash(self) -> imagehash.ImageHash:
        return imagehash.ImageHash(np.zeros([self.hash_size, self.hash_size], dtype=bool))


@dataclass(eq=True, frozen=True)
class HashOptions:
    fps: float = 5
    settings: HashSettings = DHash(8)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vidhash']

package_data = \
{'': ['*'], 'vidhash': ['stubs/*']}

install_requires = \
['ImageHash>=4.2.1,<5.0.0',
 'Pillow>=9.1.1,<10.0.0',
 'ffmpy3>=0.2.4,<0.3.0',
 'numpy>=1.22.4,<2.0.0']

setup_kwargs = {
    'name': 'vidhash',
    'version': '0.1.4',
    'description': 'A package for hashing videos and checking for similarity',
    'long_description': '# Vidhash\n[Vidhash](https://pypi.org/project/vidhash/) is a perceptual video hashing and checking library, to detect similar videos, or videos containing similar scenes.\n\n\n## Todo\n- Code\n  - Ability to ignore blank frames\n  - Ensure hash settings are the same when checking similarity\n  - Wrapper for imagehash.ImageHash\n  - Datastore\n    - (For looking up matching videos from a collection)\n- Documentation\n- Tests\n',
    'author': 'Your Name',
    'author_email': 'you@example.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/SpangleLabs/vidhash',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
